from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import Any

from acp import PROTOCOL_VERSION
from acp.exceptions import RequestError
from acp.helpers import update_available_commands, update_agent_message, text_block
from acp.schema import (
    AgentCapabilities,
    AuthenticateResponse,
    AvailableMode,
    Implementation,
    InitializeResponse,
    McpCapabilities,
    ModesResponse,
    NewSessionResponse,
    PromptResponse,
    PromptCapabilities,
)
from acp.interfaces import Agent, Client

from acp_amp.driver.base import AmpDriver
from acp_amp.mapping.to_acp import amp_event_to_updates
from acp_amp.mapping.to_amp import mcp_servers_to_amp_config, prompt_to_text


@dataclass
class _SessionState:
    session_id: str
    mode: str = "default"
    mcp_config: dict[str, Any] | None = None
    thread_id: str | None = None
    active_request_id: str | None = None
    cancelled: bool = False


class AmpAcpAgent(Agent):
    def __init__(self, driver: AmpDriver) -> None:
        self._driver = driver
        self._client: Client | None = None
        self._sessions: dict[str, _SessionState] = {}
        self._session_counter = 0

    def on_connect(self, conn: Client) -> None:
        self._client = conn

    async def initialize(
        self,
        protocol_version: int,
        client_capabilities: Any | None = None,
        client_info: Any | None = None,
        **_: Any,
    ) -> InitializeResponse:
        return InitializeResponse(
            protocol_version=PROTOCOL_VERSION,
            agent_capabilities=AgentCapabilities(
                load_session=False,
                mcp_capabilities=McpCapabilities(
                    http=True,
                    sse=True,
                ),
                prompt_capabilities=PromptCapabilities(
                    image=True,
                    embedded_context=True,
                    audio=False,
                ),
            ),
            agent_info=Implementation(
                name="acp-amp",
                title="Amp ACP Adapter",
                version="0.1.0",
            ),
            auth_methods=[],
        )

    async def authenticate(self, method_id: str, **_: Any) -> AuthenticateResponse | None:
        raise RequestError.auth_required()

    async def new_session(self, cwd: str, mcp_servers: list[Any], **_: Any) -> NewSessionResponse:
        session_id = f"amp-{self._session_counter}"
        self._session_counter += 1

        mcp_config = mcp_servers_to_amp_config(mcp_servers)
        self._sessions[session_id] = _SessionState(
            session_id=session_id,
            mcp_config=mcp_config,
        )

        if self._client:
            try:
                await self._client.session_update(
                    session_id=session_id,
                    update=update_available_commands(
                        [
                            {
                                "name": "init",
                                "description": "Generate an AGENTS.md file for the project",
                            }
                        ]
                    ),
                )
            except Exception:
                pass

        return NewSessionResponse(
            session_id=session_id,
            modes=ModesResponse(
                current_mode_id="default",
                available_modes=[
                    AvailableMode(id="default", name="Default", description="Prompt for tool permission"),
                    AvailableMode(id="bypass", name="Bypass", description="Skip tool permission prompts"),
                ],
            ),
        )

    async def load_session(self, cwd: str, mcp_servers: list[Any], session_id: str, **_: Any) -> None:
        return None

    async def set_session_mode(self, mode_id: str, session_id: str, **_: Any) -> None:
        session = self._sessions.get(session_id)
        if session:
            session.mode = mode_id
        return None

    async def prompt(self, prompt: list[Any], session_id: str, **kwargs: Any) -> PromptResponse:
        if not self._client:
            raise RuntimeError("client connection not ready")

        session = self._sessions.get(session_id)
        if not session:
            raise RuntimeError("session not found")

        session.cancelled = False
        request_id = f"{session_id}-{asyncio.get_running_loop().time()}"
        session.active_request_id = request_id

        text_prompt = prompt_to_text(prompt)
        cwd = kwargs.get("cwd")
        allow_all = session.mode == "bypass"

        stop_reason = "end_turn"
        async for item in self._driver.stream_prompt(
            prompt=text_prompt,
            cwd=cwd,
            allow_all=allow_all,
            mcp_config=session.mcp_config,
            thread_id=session.thread_id,
            request_id=request_id,
        ):
            item_type = item.get("type")
            if item_type == "event":
                event = item.get("event", {})
                thread_id = event.get("session_id") or event.get("threadId")
                if thread_id and not session.thread_id:
                    session.thread_id = str(thread_id)

                for update in amp_event_to_updates(event):
                    try:
                        await self._client.session_update(session_id=session_id, update=update)
                    except Exception:
                        pass
                continue

            if item_type == "error":
                error_text = str(item.get("error", {}).get("message", "Unknown error"))
                try:
                    await self._client.session_update(
                        session_id=session_id,
                        update=update_agent_message(text_block(f"Error: {error_text}")),
                    )
                except Exception:
                    pass
                stop_reason = "error"
                break

            if item_type == "done":
                stop_reason = item.get("stopReason") or "end_turn"
                break

        if session.cancelled:
            stop_reason = "cancelled"

        session.active_request_id = None
        session.cancelled = False

        return PromptResponse(stop_reason=stop_reason)

    async def cancel(self, session_id: str, **_: Any) -> None:
        session = self._sessions.get(session_id)
        if not session or not session.active_request_id:
            return None
        session.cancelled = True
        await self._driver.cancel(session.active_request_id)
        return None



from __future__ import annotations

import asyncio
import json
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, AsyncIterator

from acp_amp.driver.base import AmpDriver, DriverCapabilities


@dataclass
class _PendingRequest:
    queue: asyncio.Queue[dict[str, Any]]


class NodeAmpDriver(AmpDriver):
    capabilities = DriverCapabilities(
        supports_images=False,
        supports_embedded_context=True,
        supports_mcp_http=False,
        supports_mcp_sse=False,
    )

    def __init__(self, *, node_cmd: str, shim_path: Path) -> None:
        self._node_cmd = node_cmd
        self._shim_path = shim_path
        self._proc: asyncio.subprocess.Process | None = None
        self._reader_task: asyncio.Task | None = None
        self._pending: dict[str, _PendingRequest] = {}

    async def start(self) -> None:
        if self._proc:
            return
        self._proc = await asyncio.create_subprocess_exec(
            self._node_cmd,
            str(self._shim_path),
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=sys.stderr,
        )
        self._reader_task = asyncio.create_task(self._read_loop())

    async def close(self) -> None:
        if self._reader_task:
            self._reader_task.cancel()
            try:
                await self._reader_task
            except asyncio.CancelledError:
                pass
        if self._proc:
            self._proc.terminate()
            try:
                await asyncio.wait_for(self._proc.wait(), timeout=5)
            except asyncio.TimeoutError:
                self._proc.kill()
            self._proc = None

    async def stream_prompt(
        self,
        *,
        prompt: str,
        cwd: str | None,
        allow_all: bool,
        mcp_config: dict[str, Any] | None,
        thread_id: str | None,
        request_id: str,
    ) -> AsyncIterator[dict[str, Any]]:
        await self.start()
        queue: asyncio.Queue[dict[str, Any]] = asyncio.Queue()
        self._pending[request_id] = _PendingRequest(queue=queue)

        payload = {
            "type": "start",
            "id": request_id,
            "prompt": prompt,
            "cwd": cwd,
            "allowAll": allow_all,
            "mcpConfig": mcp_config or {},
            "threadId": thread_id,
        }
        await self._send(payload)

        while True:
            msg = await queue.get()
            msg_type = msg.get("type")
            if msg_type == "event":
                yield msg
                continue
            if msg_type in ("done", "error"):
                yield msg
                break

        self._pending.pop(request_id, None)

    async def cancel(self, request_id: str) -> None:
        await self.start()
        await self._send({"type": "cancel", "id": request_id})

    async def _send(self, data: dict[str, Any]) -> None:
        if not self._proc or not self._proc.stdin:
            raise RuntimeError("node shim process is not running")
        encoded = (json.dumps(data) + "\n").encode("utf-8")
        self._proc.stdin.write(encoded)
        await self._proc.stdin.drain()

    async def _read_loop(self) -> None:
        if not self._proc or not self._proc.stdout:
            return
        while True:
            line = await self._proc.stdout.readline()
            if not line:
                break
            raw = line.decode("utf-8", errors="replace").strip()
            if not raw:
                continue
            try:
                msg = json.loads(raw)
            except json.JSONDecodeError:
                continue
            request_id = str(msg.get("id", ""))
            if not request_id or request_id not in self._pending:
                continue
            await self._pending[request_id].queue.put(msg)



from __future__ import annotations

from typing import Any, Iterable

from acp.helpers import (
    image_block,
    text_block,
    tool_content,
    update_agent_message,
    update_agent_thought_text,
    update_user_message,
    start_tool_call,
    update_tool_call,
)


def amp_event_to_updates(event: dict[str, Any]) -> list[Any]:
    """Convert a single Amp SDK event into ACP session updates."""
    updates: list[Any] = []

    if event.get("type") == "result" and event.get("is_error"):
        error_text = str(event.get("error", "Unknown error"))
        updates.append(update_agent_message(text_block(f"Error: {error_text}")))
        return updates

    content = event.get("message", {}).get("content")
    if isinstance(content, str):
        updates.extend(_message_to_updates(event, [content]))
        return updates

    if isinstance(content, list):
        updates.extend(_message_to_updates(event, content))

    return updates


def _message_to_updates(event: dict[str, Any], content: Iterable[Any]) -> list[Any]:
    updates: list[Any] = []
    role = event.get("type")
    for chunk in content:
        if isinstance(chunk, str):
            updates.append(_text_update(role, chunk))
            continue

        chunk_type = chunk.get("type")
        if chunk_type == "text":
            updates.append(_text_update(role, chunk.get("text", "")))
            continue

        if chunk_type == "thinking":
            updates.append(update_agent_thought_text(chunk.get("thinking", "")))
            continue

        if chunk_type == "image":
            source = chunk.get("source", {})
            if source.get("type") == "base64":
                updates.append(
                    update_agent_message(
                        image_block(
                            data=source.get("data", ""),
                            mime_type=source.get("media_type", "application/octet-stream"),
                            uri=None,
                        )
                    )
                )
            elif source.get("type") == "url":
                updates.append(
                    update_agent_message(
                        image_block(
                            data="",
                            mime_type="",
                            uri=source.get("url"),
                        )
                    )
                )
            continue

        if chunk_type == "tool_use":
            updates.append(
                start_tool_call(
                    tool_call_id=chunk.get("id", ""),
                    title=chunk.get("name") or "Tool",
                    kind="other",
                    status="pending",
                    raw_input=_safe_json(chunk.get("input")),
                )
            )
            continue

        if chunk_type == "tool_result":
            updates.append(
                update_tool_call(
                    tool_call_id=chunk.get("tool_use_id", ""),
                    status="failed" if chunk.get("is_error") else "completed",
                    content=_tool_result_content(chunk.get("content"), is_error=bool(chunk.get("is_error"))),
                )
            )
            continue

    return updates


def _text_update(role: str | None, text: str) -> Any:
    if role == "assistant":
        return update_agent_message(text_block(text))
    return update_user_message(text_block(text))


def _tool_result_content(content: Any, *, is_error: bool) -> list[Any] | None:
    def wrap(text: str) -> str:
        if not is_error:
            return text
        return f"```\n{text}\n```"
    if isinstance(content, str) and content:
        return [tool_content(text_block(wrap(content)))]
    if isinstance(content, list) and content:
        blocks: list[Any] = []
        for item in content:
            if isinstance(item, dict) and item.get("type") == "text":
                blocks.append(tool_content(text_block(wrap(item.get("text", "")))))
        return blocks if blocks else None
    return None


def _safe_json(value: Any) -> Any:
    try:
        return jsonable(value)
    except Exception:
        return None


def jsonable(value: Any) -> Any:
    if value is None:
        return None
    if isinstance(value, (str, int, float, bool)):
        return value
    if isinstance(value, list):
        return [jsonable(v) for v in value]
    if isinstance(value, dict):
        return {str(k): jsonable(v) for k, v in value.items()}
    return str(value)



from acp_amp.mapping.to_acp import amp_event_to_updates
from acp.schema import (
    AgentMessageChunk,
    AgentThoughtChunk,
    ToolCallProgress,
    ToolCallStart,
    UserMessageChunk,
)


def _find_type(updates, cls):
    return [u for u in updates if isinstance(u, cls)]


def test_amp_text_message_to_agent_chunk():
    event = {"type": "assistant", "message": {"content": "hello"}}
    updates = amp_event_to_updates(event)
    chunks = _find_type(updates, AgentMessageChunk)
    assert chunks
    assert chunks[0].content.text == "hello"


def test_amp_thinking_to_thought_chunk():
    event = {
        "type": "assistant",
        "message": {"content": [{"type": "thinking", "thinking": "hmm"}]},
    }
    updates = amp_event_to_updates(event)
    chunks = _find_type(updates, AgentThoughtChunk)
    assert chunks
    assert chunks[0].content.text == "hmm"


def test_amp_tool_use_to_tool_call_start():
    event = {
        "type": "assistant",
        "message": {
            "content": [
                {"type": "tool_use", "id": "t1", "name": "Tool", "input": {"a": 1}}
            ]
        },
    }
    updates = amp_event_to_updates(event)
    starts = _find_type(updates, ToolCallStart)
    assert starts
    assert starts[0].tool_call_id == "t1"
    assert starts[0].status == "pending"


def test_amp_tool_result_to_tool_call_update():
    event = {
        "type": "assistant",
        "message": {
            "content": [
                {"type": "tool_result", "tool_use_id": "t1", "is_error": False, "content": "ok"}
            ]
        },
    }
    updates = amp_event_to_updates(event)
    updates_tc = _find_type(updates, ToolCallProgress)
    assert updates_tc
    assert updates_tc[0].tool_call_id == "t1"
    assert updates_tc[0].status == "completed"


def test_amp_user_text_maps_to_user_chunk():
    event = {"type": "user", "message": {"content": [{"type": "text", "text": "hi"}]}}
    updates = amp_event_to_updates(event)
    chunks = _find_type(updates, UserMessageChunk)
    assert chunks
    assert chunks[0].content.text == "hi"


def test_amp_error_result_maps_to_agent_message():
    event = {"type": "result", "is_error": True, "error": "boom"}
    updates = amp_event_to_updates(event)
    chunks = _find_type(updates, AgentMessageChunk)
    assert chunks
    assert "Error:" in chunks[0].content.text



from acp_amp.mapping.to_amp import prompt_to_text
from acp.schema import (
    EmbeddedResourceContentBlock,
    TextContentBlock,
    TextResourceContents,
)


def test_prompt_to_text_plain_text():
    prompt = [TextContentBlock(type="text", text="hello")]
    assert prompt_to_text(prompt) == "hello"


def test_prompt_to_text_init_command_expands():
    prompt = [TextContentBlock(type="text", text="/init")]
    result = prompt_to_text(prompt)
    assert "AGENTS.md" in result
    assert "Analyze this codebase" in result or "Please analyze this codebase" in result


def test_prompt_to_text_embedded_resource():
    resource = TextResourceContents(uri="file:///tmp/a.txt", text="context")
    prompt = [EmbeddedResourceContentBlock(type="resource", resource=resource)]
    result = prompt_to_text(prompt)
    assert "embedded context" in result
    assert "context" in result



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, AsyncIterator, Protocol


@dataclass
class DriverCapabilities:
    supports_images: bool = False
    supports_embedded_context: bool = True
    supports_mcp_http: bool = False
    supports_mcp_sse: bool = False


class AmpDriver(Protocol):
    capabilities: DriverCapabilities

    async def start(self) -> None:
        ...

    async def close(self) -> None:
        ...

    async def stream_prompt(
        self,
        *,
        prompt: str,
        cwd: str | None,
        allow_all: bool,
        mcp_config: dict[str, Any] | None,
        thread_id: str | None,
        request_id: str,
    ) -> AsyncIterator[dict[str, Any]]:
        ...

    async def cancel(self, request_id: str) -> None:
        ...



from __future__ import annotations

import argparse
import asyncio
import os
from pathlib import Path

from acp.core import run_agent

from acp_amp.driver.node_sdk import NodeAmpDriver
from acp_amp.server import AmpAcpAgent


def _default_shim_path() -> Path | None:
    env_path = os.environ.get("ACP_AMP_SHIM")
    if env_path:
        return Path(env_path)

    # Try repo-local layout: acp_amp/../node-shim/index.js
    here = Path(__file__).resolve()
    candidate = here.parent.parent / "node-shim" / "index.js"
    if candidate.exists():
        return candidate

    return None


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="ACP adapter for Amp Code")
    parser.add_argument(
        "--node",
        default=os.environ.get("ACP_AMP_NODE", "node"),
        help="Node.js executable (default: node)",
    )
    parser.add_argument(
        "--shim",
        default=str(_default_shim_path() or ""),
        help="Path to the Node shim (default: auto-detect or ACP_AMP_SHIM)",
    )
    return parser.parse_args()


async def _run() -> None:
    args = _parse_args()
    if not args.shim:
        raise SystemExit("ACP AMP: node shim not found; set ACP_AMP_SHIM or --shim")

    driver = NodeAmpDriver(node_cmd=args.node, shim_path=Path(args.shim))
    agent = AmpAcpAgent(driver)
    await run_agent(agent)


def main() -> None:
    asyncio.run(_run())


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Any

from acp.schema import (
    EmbeddedResourceContentBlock,
    ImageContentBlock,
    ResourceContentBlock,
    TextContentBlock,
)


def prompt_to_text(prompt: list[Any]) -> str:
    parts: list[str] = []
    for block in prompt:
        if isinstance(block, TextContentBlock):
            text = block.text
            if text.strip() == "/init":
                text = _init_instructions()
            parts.append(text)
            continue
        if isinstance(block, ResourceContentBlock):
            parts.append(f"\n[resource link] {block.uri}\n")
            continue
        if isinstance(block, EmbeddedResourceContentBlock):
            resource = block.resource
            text = getattr(resource, "text", None)
            uri = getattr(resource, "uri", "")
            if text is not None:
                parts.append(f"\n[embedded context {uri}]\n{text}\n[context end]\n")
            continue
        if isinstance(block, ImageContentBlock):
            parts.append("\n[image omitted]\n")
            continue
    return "".join(parts)


def _init_instructions() -> str:
    return (
        "Please analyze this codebase and create an AGENTS.md file containing:\n"
        "1. Build/lint/test commands - especially for running a single test\n"
        "2. Architecture and codebase structure information, including important subprojects, "
        "internal APIs, databases, etc.\n"
        "3. Code style guidelines, including imports, conventions, formatting, types, naming "
        "conventions, error handling, etc.\n\n"
        "The file you create will be given to agentic coding tools (such as yourself) that "
        "operate in this repository. Make it about 20 lines long.\n\n"
        "If there are Cursor rules (in .cursor/rules/ or .cursorrules), Claude rules (CLAUDE.md), "
        "Windsurf rules (.windsurfrules), Cline rules (.clinerules), Goose rules (.goosehints), "
        "or Copilot rules (in .github/copilot-instructions.md), make sure to include them. "
        "Also, first check if there is an existing AGENTS.md or AGENT.md file, and if so, "
        "update it instead of overwriting it."
    )


def mcp_servers_to_amp_config(mcp_servers: list[Any]) -> dict[str, Any]:
    mcp_config: dict[str, Any] = {}
    for server in mcp_servers:
        name = getattr(server, "name", None)
        command = getattr(server, "command", None)
        args = getattr(server, "args", None)
        env = getattr(server, "env", None)
        if not name or not command:
            continue
        env_map = None
        if env:
            env_map = {e.name: e.value for e in env}
        mcp_config[name] = {
            "command": command,
            "args": args or [],
            "env": env_map,
        }
    return mcp_config
